@@ -10,7 +10,7 @@
  * redistribution of this file, and for a DISCLAIMER OF ALL
  * WARRANTIES.
  *
- * RCS: @(#) $Id: FuncSig.java,v 1.5 1999/05/15 23:38:14 dejong Exp $
+ * RCS: @(#) $Id: FuncSig.java,v 1.15 2006/05/15 22:14:23 mdejong Exp $
  *
  */
 
@@ -28,7 +28,7 @@ import java.util.*;
  * signature."
  */
 
-class FuncSig extends InternalRep {
+class FuncSig implements InternalRep {
 
 // The class that a method signature is used against. In the case of a
 // static method call by java::call, targetCls is given by the <class>
@@ -72,9 +72,13 @@ PkgInvoker pkgInvoker;
 
 Object func;
 
-// Stores all of the declared methods of each Java class.
+// Stores all accessible instance methods for a Java class
+// Note that we use a Hashtable instead of a HashMap here
+// since these fields could be accessed from multiple
+// threads and the Hashtable class is synchronized.
 
-static Hashtable allDeclMethTable = new Hashtable();
+static Hashtable instanceMethodTable = new Hashtable();
+static Hashtable staticMethodTable = new Hashtable();
 
 
 /*
@@ -116,11 +120,17 @@ FuncSig(
  *----------------------------------------------------------------------
  */
 
-protected InternalRep
-duplicate()
+public InternalRep duplicate()
 {
     return new FuncSig(targetCls, pkgInvoker, func);
 }
+
+/**
+  * Implement this no-op for the InternalRep interface.
+  */
+
+public void dispose() {}
+
 
 /*
  *----------------------------------------------------------------------
@@ -131,11 +141,12 @@ duplicate()
  *	or method that matches with the signature and the parameters.
  *
  * Results:
- *	The FuncSig given by the signature.
+ *	The FuncSig given by the signature. An exception will be
+ *	raised if the constructor provided is for an abstract
+ *	class, or if the class is not accessible.
  *
  * Side effects:
- *	When successful, the internalRep of the signature object is
- *	converted to FuncSig.
+ *	None.
  *
  *----------------------------------------------------------------------
  */
@@ -149,8 +160,9 @@ get(
     TclObject signature,	// Method/constructor signature.
     TclObject[] argv,		// Arguments.
     int startIdx,		// Index of the first argument in argv
-    int count)			// Number of arguments to pass to the
+    int count,			// Number of arguments to pass to the
 				// constructor.
+    boolean isStatic)		// True if signature is for a static Method.
 throws
     TclException
 {
@@ -194,16 +206,30 @@ throws
     Object match;
     int sigLength = TclList.getLength(interp, signature);
     String methodName = null;
+    TclObject class_or_method;
 
     if (sigLength == 0) {
 	throw new TclException(interp, "bad signature \"" + signature + "\"");
+    } else if (sigLength == 1) {
+	class_or_method = signature;
+    } else {
+	class_or_method = TclList.index(interp, signature, 0);
     }
 
     if (isConstructor) {
-	cls = JavaInvoke.getClassByName(interp,
-		  TclList.index(interp,signature, 0).toString() );
+	cls = JavaInvoke.getClassByName(interp, class_or_method.toString() );
     } else {
-	methodName = TclList.index(interp, signature, 0).toString();
+	methodName = class_or_method.toString();
+    }
+
+    if ((isConstructor || isStatic) && !PkgInvoker.isAccessible(cls)) {
+        JavaInvoke.notAccessibleError(interp, cls);
+    }
+
+    if (isConstructor && Modifier.isAbstract(cls.getModifiers())) {
+        throw new TclException(interp, "Class \"" +
+	    JavaInfoCmd.getNameFromClass(cls) +
+	    "\" is abstract");
     }
 
     if ((sigLength > 1) || (sigLength == 1 && count == 0)) {
@@ -226,27 +252,29 @@ throws
 
 	if (isConstructor) {
 	    try {
-		match = cls.getDeclaredConstructor(paramTypes);
+		match = getAccessibleConstructor(cls, paramTypes);
 	    } catch (NoSuchMethodException e) {
 	      if (sigLength > 1) {
-		throw new TclException(interp, "no such constructor \"" +
+		throw new TclException(interp, "no accessible constructor \"" +
 			    signature + "\"");
 	      } else {
-		throw new TclException(interp, "can't find constructor with " +
-			    count + " argument(s) for class \"" + cls.getName() +
+		throw new TclException(interp, "can't find accessible constructor with " +
+			    count + " argument(s) for class \"" +
+			      JavaInfoCmd.getNameFromClass(cls) +
 			      "\"");
 	      }
 	    }
 	} else {
-	  match = lookupMethod(interp, cls, methodName, paramTypes, signature);
+	  match = lookupMethod(interp, cls, methodName, paramTypes,
+	          signature, isStatic);
 	}
     } else {
       match = matchSignature(interp, cls, signature, methodName,
-			     isConstructor, argv, startIdx, count);
+			     isConstructor, argv, startIdx, count, isStatic);
     }
 
     FuncSig sig = new FuncSig(cls, PkgInvoker.getPkgInvoker(cls), match);
-    signature.setInternalRep(sig);
+    //signature.setInternalRep(sig);
 
     return sig;
 }
@@ -258,7 +286,7 @@ throws
 
 
 // lookupMethod attempts to find an exact match for the method name
-// based on the typs (Java Class objects) of the arguments to the
+// based on the types (Java Class objects) of the arguments to the
 // method. If an exact match can not be found it will raise a TclException.
 
 static Method
@@ -267,13 +295,19 @@ lookupMethod(
     Class cls,               // the Java objects class
     String methodName,       // name of method
     Class[] paramTypes,      // the Class object arguments
-    TclObject  signature     // used for error reporting
+    TclObject  signature,    // used for error reporting
+    boolean isStatic         // True if signature is for a static Method.
 )
     throws TclException
 {
-  Method[] methods = getAllDeclaredMethods(cls);
+  Method[] methods;
   boolean foundSameName = false;
 
+  if (isStatic)
+      methods = getAccessibleStaticMethods(cls);
+  else
+      methods = getAccessibleInstanceMethods(cls);
+
 
   // FIXME : searching Java methods for method name match
   // searching through ALL the methods is really slow
@@ -307,13 +341,16 @@ lookupMethod(
 
   if (paramTypes.length > 0 || !foundSameName) {
     throw new TclException(interp,
-      "no such method \"" + signature + "\" in class " +
-			   cls.getName());
+      "no accessible" + (isStatic?" static ":" ") + "method \"" +
+      signature + "\" in class " +
+      JavaInfoCmd.getNameFromClass(cls));
   } else {
     throw new TclException(interp,
-      "can't find method \"" + signature + "\" with " +
-        paramTypes.length + " argument(s) for class \"" + cls.getName()
-			   + "\"");
+      "can't find accessible" + (isStatic?" static ":" ") + "method \"" +
+      signature + "\" with " + paramTypes.length +
+      " argument(s) for class \"" +
+      JavaInfoCmd.getNameFromClass(cls) +
+      "\"");
   }
 }
 
@@ -328,7 +365,7 @@ lookupMethod(
 // This method will attempt to find a match for a signature
 // if an exact match can not be found then it will use
 // the types of the argument objects to "guess" what
-// method was intended by the user. If not match can be
+// method was intended by the user. If no match can be
 // found after "guessing" then a TclException will be raised.
 
 
@@ -336,26 +373,30 @@ static Object
 matchSignature(
     Interp interp,           // the tcl interpreter
     Class cls,               // the Java objects class
-    TclObject  signature,    // used for error reporting
+    TclObject signature,     // used for error reporting
     String methodName,       // name of method, can be null
     boolean isConstructor,   // duh
     TclObject[] argv,        // arguments to Method or Constructor
     int startIdx,	     // Index of the first argument in argv    
-    int argv_count           // set to -1 if JFK was killed by the FBI
+    int argv_count,          // set to -1 if JFK was killed by the FBI
+    boolean isStatic         // True if signature is for a static Method.
 )
     throws TclException
 {
   Object[] funcs;
   boolean foundSameName = false;
-  Vector match_vector = new Vector();
+  ArrayList match_list = new ArrayList();
   int i,j;
 
   final boolean debug = false;  
 
   if (isConstructor) {
-    funcs = cls.getDeclaredConstructors();
+    funcs = getAccessibleConstructors(cls);
   } else {
-    funcs = getAllDeclaredMethods(cls);
+    if (isStatic)
+        funcs = getAccessibleStaticMethods(cls);
+    else
+        funcs = getAccessibleInstanceMethods(cls);
   }
 
   for (i = 0; i < funcs.length; i++) {
@@ -373,19 +414,19 @@ matchSignature(
     }
      
     if (paramTypes.length == argv_count) {
-      match_vector.addElement(funcs[i]);
+      match_list.add(funcs[i]);
     }
   }  
 
 
   // If there is only a single remaining match then we can return it now
 
-  if (match_vector.size() == 1) {
+  if (match_list.size() == 1) {
      // debug print the single match
-     //System.out.println("single match : " + match_vector.elementAt(0));
+     //System.out.println("single match : " + match_list.get(0));
 
-    return match_vector.elementAt(0);
-  } else if (match_vector.size() > 1) {
+    return match_list.get(0);
+  } else if (match_list.size() > 1) {
     
     Class[] argv_classes = new Class[argv_count];
     Class[] match_classes;
@@ -431,12 +472,12 @@ matchSignature(
 
     // debug print possible match types
 
-    for (i=0; i < match_vector.size(); i++) {
+    for (i=0; i < match_list.size(); i++) {
 
       if (isConstructor) {
-	match_classes = ((Constructor) match_vector.elementAt(i)).getParameterTypes();
+	match_classes = ((Constructor) match_list.get(i)).getParameterTypes();
       } else {
-	match_classes = ((Method) match_vector.elementAt(i)).getParameterTypes();
+	match_classes = ((Method) match_list.get(i)).getParameterTypes();
       }
 
       System.out.print("match " + i + " is ");
@@ -456,12 +497,12 @@ matchSignature(
     // try to match the argument types and the
     // match types exactly by comparing Class objects
     
-    for (i=0; i < match_vector.size(); i++) {
+    for (i=0; i < match_list.size(); i++) {
 
       if (isConstructor) {
-	match_classes = ((Constructor) match_vector.elementAt(i)).getParameterTypes();
+	match_classes = ((Constructor) match_list.get(i)).getParameterTypes();
       } else {
-	match_classes = ((Method) match_vector.elementAt(i)).getParameterTypes();
+	match_classes = ((Method) match_list.get(i)).getParameterTypes();
       }
 
       boolean exact = true;
@@ -476,14 +517,14 @@ matchSignature(
         if (debug) {
 	System.out.println("exact match at " + i);
         } // end if (debug)
-	return match_vector.elementAt(i);
+	return match_list.get(i);
       }
     }
 
 
 
 
-    // loop from the end of the Vector to the begining and
+    // loop from the end of the list to the begining and
     // remove those signatures that are not assignable
     // take special care not to remove signatures that
     // have an object decended from java.lang.Object
@@ -491,34 +532,25 @@ matchSignature(
     // This means a null argument will not match a built in type.
     
 
-    for (i = match_vector.size()-1; i >= 0 ; i--) {
+    for (i = match_list.size()-1; i >= 0 ; i--) {
       
       if (isConstructor) {
-	match_classes = ((Constructor) match_vector.elementAt(i)).getParameterTypes();
+	match_classes = ((Constructor) match_list.get(i)).getParameterTypes();
       } else {
-	match_classes = ((Method) match_vector.elementAt(i)).getParameterTypes();
+	match_classes = ((Method) match_list.get(i)).getParameterTypes();
       }
 
-      // Test for assignability.  A primitive type can not be passed a null
-      // argument so we check for that case first. If the argument is
-      // null we do not throw out a non primitive argument as any Class
-      // could be assigned the null object. Then we check for assignability
-      // of the Class object itself and throw out any that do not jive.
+      // If any of the arguments are not assignable to the method arguments,
+      // then remove this method from the list of matches.
 
       for (j=0; j < argv_count; j++) {
-	if ((match_classes[j].isPrimitive() && argv_classes[j] == null) ||
-            (argv_classes[j] != null &&
-            match_classes[j] != argv_classes[j] &&
-	    !match_classes[j].isAssignableFrom(argv_classes[j]))) {
-	  
-	  // if this argument is totally incompatible with the signature
-	  // then remove this signature from the possible matches
+	if (!JavaInvoke.isAssignable(match_classes[j], argv_classes[j])) {
 	  
           if (debug) {
           System.out.println("removing non assignable match " + i);
           } // end if (debug)
 
-	  match_vector.removeElementAt(i);
+	  match_list.remove(i);
 	  break; // go on to next Method
 	}
       }
@@ -528,18 +560,18 @@ matchSignature(
 
 
     if (debug) {
-    // debug print match_vector after isAssignabelFrom test
+    // debug print match_list after isAssignabelFrom test
 
-    if (match_vector.size() > 0) {
+    if (match_list.size() > 0) {
       System.out.println("isAssignableFrom() matches");
     }
 
-    for (i=0; i < match_vector.size(); i++) {
+    for (i=0; i < match_list.size(); i++) {
 
       if (isConstructor) {
-	match_classes = ((Constructor) match_vector.elementAt(i)).getParameterTypes();
+	match_classes = ((Constructor) match_list.get(i)).getParameterTypes();
       } else {
-	match_classes = ((Method) match_vector.elementAt(i)).getParameterTypes();
+	match_classes = ((Method) match_list.get(i)).getParameterTypes();
       }
 
       System.out.print("match " + i + " is ");
@@ -557,21 +589,21 @@ matchSignature(
 
     // If there is only a single remaining match then we can return it now
 
-    if (match_vector.size() == 1) {
-      return match_vector.elementAt(0);
+    if (match_list.size() == 1) {
+      return match_list.get(0);
     }
 
   
 
 
-    // at this point match_vector should have only those signatures
+    // at this point match_list should have only those signatures
     // that can take the argument types from argv with widining conversion
     
     // to figure out which method we should call we need to determine
     // which signatures are "better" then others where "better" is defined
     // as the shortest number of steps up the inheritance or interface tree
     
-    if (match_vector.size() > 1) {
+    if (match_list.size() > 1) {
     
       // the first thing we need to do is get the inheritance info
       // of the arguments to the method. From this we will create
@@ -598,7 +630,7 @@ matchSignature(
       // we use a vector to store up all of the Class objects
       // that make up the inheritance tree for a particular class
 
-      Vector class_vector = new Vector();
+      ArrayList class_list = new ArrayList();
 
 
       // for each argument to the method we loop up the inheritance tree
@@ -632,25 +664,26 @@ matchSignature(
 
 	while (c != null) {
 	  // add a null to the front of the vector
-	  class_vector.addElement(null);
+	  class_list.add(null);
 
 	  // add this Class and its Interfaces to the vector
-	  addInterfaces(c,class_vector);
+	  addInterfaces(c, class_list);
 
 	  c = c.getSuperclass();
 	}
 
 
 	// now remove the first element of the vector (it is null)
-	class_vector.removeElementAt(0);
+	class_list.remove(0);
 
-	Class[] classes = new Class[class_vector.size()];
-	
-	class_vector.copyInto(classes);
+	Class[] classes = new Class[class_list.size()];
+	for (j=0; j < classes.length; j++) {
+            classes[j] = (Class) class_list.get(j);
+        }
 
 	argv_classes_lookup[i] = classes;
 
-	class_vector.removeAllElements();
+	class_list.clear();
       }
       
       
@@ -692,9 +725,9 @@ matchSignature(
 
 
 
-      int[] super_steps = new int[match_vector.size()];
-      int[] total_steps = new int[match_vector.size()];
-      boolean[] trim_matches = new boolean[match_vector.size()];
+      int[] super_steps = new int[match_list.size()];
+      int[] total_steps = new int[match_list.size()];
+      boolean[] trim_matches = new boolean[match_list.size()];
       int min_super_step;
       int min_total_step;
       Class min_class;
@@ -721,12 +754,12 @@ matchSignature(
 	// iterate over the matched methods to find the
 	// minimum steps for this argument
 	
-	for (i=0; i < match_vector.size(); i++) {
+	for (i=0; i < match_list.size(); i++) {
 
 	  if (isConstructor) {
-	    match_classes = ((Constructor) match_vector.elementAt(i)).getParameterTypes();
+	    match_classes = ((Constructor) match_list.get(i)).getParameterTypes();
 	  } else {
-	    match_classes = ((Method) match_vector.elementAt(i)).getParameterTypes();
+	    match_classes = ((Method) match_list.get(i)).getParameterTypes();
 	  }
 
 	  Class match_to = match_classes[j];
@@ -805,7 +838,7 @@ matchSignature(
 	
 	System.out.println("step arrays for argument " + j);
 
-	for (int loop=0; loop < match_vector.size(); loop++) {
+	for (int loop=0; loop < match_list.size(); loop++) {
 	  System.out.println("(" + super_steps[loop] + "," + total_steps[loop] + ")");
 	}
 
@@ -818,7 +851,7 @@ matchSignature(
 	// from the step info we know the minumum so we can
 	// remove those values that are "worse" then the min
 
-	for (i=match_vector.size()-1; i >= 0; i--) {
+	for (i=match_list.size()-1; i >= 0; i--) {
 
 	  if (super_steps[i] > min_super_step ||
 	      (super_steps[i] == min_super_step &&
@@ -849,9 +882,9 @@ matchSignature(
 
       // remove the methods that were marked for deletion
       
-      for (i=match_vector.size()-1; i >= 0; i--) {
+      for (i=match_list.size()-1; i >= 0; i--) {
 	if (trim_matches[i]) {
-	  match_vector.removeElementAt(i);
+	  match_list.remove(i);
 	}
       }
 
@@ -864,12 +897,12 @@ matchSignature(
       
       System.out.println("after super steps trim");
       
-      for (i=0; i < match_vector.size(); i++) {
+      for (i=0; i < match_list.size(); i++) {
 	
 	if (isConstructor) {
-	  match_classes = ((Constructor) match_vector.elementAt(i)).getParameterTypes();
+	  match_classes = ((Constructor) match_list.get(i)).getParameterTypes();
 	} else {
-	  match_classes = ((Method) match_vector.elementAt(i)).getParameterTypes();
+	  match_classes = ((Method) match_list.get(i)).getParameterTypes();
 	}
 	
 	System.out.print("match " + i + " is ");
@@ -885,20 +918,20 @@ matchSignature(
       } // end if (debug)
      
       
-    } // end if (match_vector.size() > 1)
+    } // end if (match_list.size() > 1)
  
     
     
-    // if there is only one item left in the match_vector return it
+    // if there is only one item left in the match_list return it
     
-    if (match_vector.size() == 1) {
-      return match_vector.elementAt(0);
+    if (match_list.size() == 1) {
+      return match_list.get(0);
     } else {
       // if we have 0 or >1 remaining matches then
       // we were unable to find the "best" match so raise an error
       // if possible, tell user what matches made the sig ambiguous.
 
-      //System.out.println("match_vector.size() is " + match_vector.size());
+      //System.out.println("match_list.size() is " + match_list.size());
 
 
       StringBuffer sb = new StringBuffer(100);
@@ -911,7 +944,7 @@ matchSignature(
       sb.append(" signature");
 
 
-      if (match_vector.size() == 0) {
+      if (match_list.size() == 0) {
 
 	  // FIXME : better error message for no signature matches case
 	  // We really should tell the user which methods we could have
@@ -925,9 +958,12 @@ matchSignature(
           // Get all the signatures that match this name and number or args
 
           if (isConstructor) {
-            funcs = cls.getDeclaredConstructors();
+            funcs = getAccessibleConstructors(cls);
           } else {
-            funcs = getAllDeclaredMethods(cls);
+            if (isStatic)
+              funcs = getAccessibleStaticMethods(cls);
+            else
+              funcs = getAccessibleInstanceMethods(cls);
           }
 
           for (i = 0; i < funcs.length; i++) {
@@ -945,7 +981,7 @@ matchSignature(
             }
      
             if (paramTypes.length == argv_count) {
-              match_vector.addElement(funcs[i]);
+              match_list.add(funcs[i]);
             }
           }  
 
@@ -963,12 +999,12 @@ matchSignature(
       TclObject siglist = TclList.newInstance();
       siglist.preserve();
 
-      for (i=0; i < match_vector.size(); i++) {
+      for (i=0; i < match_list.size(); i++) {
           TclObject cur_siglist = TclList.newInstance();
           cur_siglist.preserve();
 
           if (isConstructor) {
-            Constructor con = ((Constructor) match_vector.elementAt(i));
+            Constructor con = ((Constructor) match_list.get(i));
             TclList.append(interp,cur_siglist,
                 TclString.newInstance(con.getName()) );
 
@@ -976,7 +1012,7 @@ matchSignature(
 
 	    match_classes = con.getParameterTypes();
 	  } else {
-            Method meth = ((Method) match_vector.elementAt(i));
+            Method meth = ((Method) match_list.get(i));
             TclList.append(interp,cur_siglist,
                 TclString.newInstance(meth.getName()) );
 
@@ -1003,7 +1039,7 @@ matchSignature(
       throw new TclException(interp, sb.toString());
     }
     
-  } // end else if (match_vector.size() > 1)
+  } // end else if (match_list.size() > 1)
   
 
 
@@ -1011,20 +1047,23 @@ matchSignature(
   // if we got to here then we could not find a matching method so raise error
 
   if (isConstructor) {
-    throw new TclException(interp, "can't find constructor with " +
+    throw new TclException(interp, "can't find accessible constructor with " +
 			       argv_count + " argument(s) for class \"" +
-			       cls.getName() + "\"");
+			       JavaInfoCmd.getNameFromClass(cls) +
+			       "\"");
   } else {
-    
     if (!foundSameName) {
       throw new TclException(interp,
-			     "no such method \"" + signature + "\" in class " +
-			     cls.getName());
+			     "no accessible" + (isStatic?" static ":" ") +
+			     "method \"" + signature + "\" in class " +
+			     JavaInfoCmd.getNameFromClass(cls));
     } else {
       throw new TclException(interp,
-			     "can't find method \"" + signature + "\" with " +
+			     "can't find accessible" + (isStatic?" static ":" ") +
+			     "method \"" + signature + "\" with " +
 			     argv_count + " argument(s) for class \"" +
-			     cls.getName() + "\"");
+			     JavaInfoCmd.getNameFromClass(cls) +
+			     "\"");
     }
   }
 }
@@ -1036,36 +1075,125 @@ matchSignature(
 
 // Helper method to recursively add interfaces of a class to the vector
 
-private static void addInterfaces(Class cls, Vector v)
+private static void addInterfaces(Class cls, ArrayList alist)
 {
-  v.addElement(cls);
+  alist.add(cls);
 
   Class[] interfaces = cls.getInterfaces();
-  
+
   for (int i=0; i < interfaces.length ; i++) {
-    addInterfaces(interfaces[i],v);
+    addInterfaces(interfaces[i], alist);
   }
 }
 
 
 
 
-
 
 /*
  *----------------------------------------------------------------------
  *
- * getAllDeclaredMethods --
+ * getAccessibleConstructors --
  *
- *	Returns all the methods declared by the class or superclasses
- *	of the class.
+ *	Returns all constructors that can be invoked for a given class.
  *
  * Results:
- *	An array of all the methods declared by the class and the
- *	superclasses of this class. If overloaded methods, only the
+ *	An array of all the accessible constructors in the class.
+ *
+ * Side effects:
+ *	None.
+ *
+ *----------------------------------------------------------------------
+ */
+
+static Constructor[]
+getAccessibleConstructors(
+    Class cls)				// The class to query.
+{
+    if (PkgInvoker.usesDefaultInvoker(cls)) {
+	return cls.getConstructors();
+    } else {
+	Constructor[] constructors = cls.getDeclaredConstructors();
+	ArrayList alist = null;
+	boolean skipped_any = false;
+
+	for (int i=0; i < constructors.length; i++) {
+	    Constructor c = constructors[i];
+	    if (PkgInvoker.isAccessible(c)) {
+	        if (alist == null) {
+	            alist = new ArrayList(constructors.length);
+	        }
+	        alist.add(c);
+	    } else {
+	        skipped_any = true;
+	    }
+	}
+
+	if (skipped_any) {
+            if (alist == null) {
+                constructors = new Constructor[0];
+            } else {
+	        constructors = new Constructor[alist.size()];
+	        for (int i=0; i < constructors.length; i++) {
+                    constructors[i] = (Constructor) alist.get(i);
+                }
+            }
+	}
+	return constructors;
+    }
+}
+
+/*
+ *----------------------------------------------------------------------
+ *
+ * getAccessibleConstructor --
+ *
+ *	Returns an accessable constructors for the given class
+ *	that accepts the given arguments.
+ *
+ * Results:
+ *	A constructor object, raises a NoSuchMethodException
+ *	if an accessible constructor cannot be found.
+ *
+ * Side effects:
+ *	None.
+ *
+ *----------------------------------------------------------------------
+ */
+
+static Constructor
+getAccessibleConstructor(
+    Class cls,				// The class to query.
+    Class[] parameterTypes)		// The constructor arguments types
+    throws NoSuchMethodException
+{
+    if (PkgInvoker.usesDefaultInvoker(cls)) {
+	return cls.getConstructor(parameterTypes);
+    } else {
+	Constructor constructor =
+	    cls.getDeclaredConstructor(parameterTypes);
+	if (!PkgInvoker.isAccessible(constructor)) {
+	    throw new NoSuchMethodException();
+	}
+	return constructor;
+    }
+}
+
+/*
+ *----------------------------------------------------------------------
+ *
+ * getAccessibleInstanceMethods --
+ *
+ *	Returns all instance methods that can be invoked for a given class.
+ *
+ * Results:
+ *	An array of all the accessible instance methods in the class and the
+ *	superclasses of the class. If a method is overloaded, only the
  *	"most public" instance of that method is included in the
- *	array. See comments above the "func" member variable for more
- *	details.
+ *	array. A method is considered accessible if it has public
+ *	access or if it does not have private access and the package has
+ *	a custom PkgInvoker. See comments above the "func" member
+ *	variable for more details.
  *
  * Side effects:
  *	The array of methods are saved in a hashtable for faster access
@@ -1075,36 +1203,42 @@ private static void addInterfaces(Class cls, Vector v)
  */
 
 static Method[]
-getAllDeclaredMethods(
+getAccessibleInstanceMethods(
     Class cls)				// The class to query.
 {
-    Method[] methods = (Method[]) allDeclMethTable.get(cls);
+    Method[] methods = (Method[]) instanceMethodTable.get(cls);
     if (methods != null) {
 	return methods;
     }
 
-    Vector vec = new Vector();
+    // Avoid using Class.getMethods() because it includes
+    // static members and it does not account for interfaces
+    // which need to include methods from the Object class.
+
+    ArrayList alist = new ArrayList();
 
     for (Class c = cls; c != null; ) {
-	mergeMethods(c, c.getDeclaredMethods(), vec);
+	methods = c.getDeclaredMethods();
+	mergeInstanceMethods(c, methods, alist);
 
 	Class interfaces[] = c.getInterfaces();
 	for (int i = 0; i < interfaces.length; i++) {
-	    mergeMethods(interfaces[i], interfaces[i].getMethods(), vec);
+	    mergeInstanceMethods(interfaces[i], interfaces[i].getMethods(), alist);
 	}
 
 	if (c.isInterface()) {
 	    c = Object.class; // if cls is an interface add Object methods
 	} else {
 	    c = c.getSuperclass();
-	}	
-
+	}
     }
 
-    sortMethods(vec); // removed until this can be tested more.
-    methods = new Method[vec.size()];
-    vec.copyInto(methods);
-    allDeclMethTable.put(cls, methods);
+    sortMethods(alist);
+    methods = new Method[alist.size()];
+    for (int i=0; i < methods.length; i++) {
+        methods[i] = (Method) alist.get(i);
+    }
+    instanceMethodTable.put(cls, methods);
 
     return methods;
 }
@@ -1112,10 +1246,66 @@ getAllDeclaredMethods(
 /*
  *----------------------------------------------------------------------
  *
- * mergeMethods --
+ * getAccessibleStaticMethods --
  *
- *	Add the methods declared by a super-class or an interface to
- *	the list of declared methods of a class.
+ *	Returns all static methods that can be invoked for a given class.
+ *
+ * Results:
+ *	An array of all the accessible static methods in the class.
+ *	A method is considered accessible if it has public
+ *	access or if it is not private and the package has
+ *	a custom PkgInvoker. See comments above the "func" member
+ *	variable for more details.
+ *
+ * Side effects:
+ *	None.
+ *
+ *----------------------------------------------------------------------
+ */
+
+static Method[]
+getAccessibleStaticMethods(
+    Class cls)				// The class to query.
+{
+    Method[] methods = (Method[]) staticMethodTable.get(cls);
+    if (methods != null) {
+	return methods;
+    }
+
+    // When searching for static methods in this class,
+    // call getDeclaredMethods() and filter out those
+    // methods that are not static or are not accessible.
+    // This should be quicker than calling getMethods()
+    // since that returns both static and instance methods
+    // for the class and its superclasses.
+
+    methods = cls.getDeclaredMethods();
+    ArrayList alist = new ArrayList();
+
+    for (int i=0; i < methods.length; i++) {
+	Method m = methods[i];
+	if (Modifier.isStatic(m.getModifiers()) && PkgInvoker.isAccessible(m)) {
+	    alist.add(m);
+	}
+    }
+
+    sortMethods(alist);
+    methods = new Method[alist.size()];
+    for (int i=0; i < methods.length; i++) {
+        methods[i] = (Method) alist.get(i);
+    }
+    staticMethodTable.put(cls, methods);
+
+    return methods;
+}
+
+/*
+ *----------------------------------------------------------------------
+ *
+ * mergeInstanceMethods --
+ *
+ *	Add instance methods declared by a super-class or an interface to
+ *	the list of accessible instance methods.
  *
  * Results:
  *	None.
@@ -1129,17 +1319,26 @@ getAllDeclaredMethods(
  */
 
 private static void 
-mergeMethods(
+mergeInstanceMethods(
     Class c,
     Method methods[],
-    Vector vec)
+    ArrayList alist)
 {
     for (int i=0; i < methods.length; i++) {
 	boolean sameSigExists = false;
 	Method newMeth = methods[i];
 
-	for (int j=0; j < vec.size(); j++) {
-	    Method oldMeth = (Method)vec.elementAt(j);
+	if (newMeth == null)
+	    continue;
+
+	// Don't merge static methods or inaccessible methods
+	if (Modifier.isStatic(newMeth.getModifiers()) ||
+	        !PkgInvoker.isAccessible(newMeth)) {
+	    continue;
+	}
+
+	for (int j=0; j < alist.size(); j++) {
+	    Method oldMeth = (Method) alist.get(j);
 
 	    if (methodSigEqual(oldMeth, newMeth)) {
 		sameSigExists = true;
@@ -1149,21 +1348,21 @@ mergeMethods(
 		int oldRank = getMethodRank(oldCls, oldMeth);
 
 		if (newRank > oldRank) {
-		    vec.setElementAt(newMeth, j);
+		    alist.set(j, newMeth);
 		}
 		break;
 	    }
 	}
 	    
 	if (!sameSigExists) {
-	    // We copy a method into the vector only if no method
-	    // with the same signature is already in the
-	    // vector. Otherwise the matching routine in the get()
+	    // We copy a method into the list only if no method
+	    // with the same signature is already in the list.
+	    // Otherwise the matching routine in the get()
 	    // procedure may run into "ambiguous method signature"
 	    // errors when it sees instances of overloaded
 	    // methods.
 
-	    vec.addElement(newMeth);
+	    alist.add(newMeth);
 	}
     }
 }
@@ -1215,43 +1414,43 @@ methodSigEqual(
  *
  * sortMethods --
  *
- *	This method will sort a vector of Method objects. We need to sort
+ *	This method will sort a list of Method objects. We need to sort
  *      the methods so that the order of the methods does not depend on
- *      the order the methods are returned by the JVM.
+ *      the order the methods as returned by the JVM.
  *
  * Results:
  *	None.
  *
  * Side effects:
- *      The order of the elements in the Vector is changed.
+ *      The order of the elements in the List is changed.
  *
  *----------------------------------------------------------------------
  */
 
 private static void 
 sortMethods(
-    Vector vec)
+    ArrayList alist)
 {
     final boolean debug = false; // set to true for debug output
-    int insize = vec.size();
+    int insize = alist.size();
 
     if (debug) {
 	System.out.println("Pre sort dump");
-	for (int i=0; i < vec.size(); i++) {
-	    Method m = (Method) vec.elementAt(i);
+	for (int i=0; i < alist.size(); i++) {
+	    Method m = (Method) alist.get(i);
 	    System.out.println("Method " + i + " is \t\"" + getMethodDescription(m) + "\"");
 	}
     }
 
-    for (int i=1; i < vec.size(); i++) {
+    for (int i=1; i < alist.size(); i++) {
 	int c = i;
-	Method cm = (Method) vec.elementAt(c);
+	Method cm = (Method) alist.get(c);
 	String cms = getMethodDescription(cm);
 
 	// loop down array swapping elements into sorted order
 
 	for (int j = c - 1; j >= 0 ; j--) {
-	    Method jm = (Method) vec.elementAt(j);
+	    Method jm = (Method) alist.get(j);
 	    String jms = getMethodDescription(jm);
 
 	    if (debug) {
@@ -1266,8 +1465,8 @@ sortMethods(
 		}
 		
 		// Swap the Methods at c and j
-		vec.setElementAt(jm,c);
-	    	vec.setElementAt(cm,j);
+		alist.set(c, jm);
+	    	alist.set(j, cm);
 
 		// Current becomes the value of j for next loop
 		c = j;
@@ -1285,13 +1484,13 @@ sortMethods(
 
     if (debug) {
 	System.out.println("Post sort dump");
-	for (int i=0; i < vec.size(); i++) {
-	    Method m = (Method) vec.elementAt(i);
+	for (int i=0; i < alist.size(); i++) {
+	    Method m = (Method) alist.get(i);
 	    System.out.println("Method " + i + " is \t\"" +  getMethodDescription(m)+ "\"");
 	}
     }
 
-    if (insize != vec.size()) {
+    if (insize != alist.size()) {
 	throw new RuntimeException("lost elements");
     }
     return;
